"""
Class: ConfigurationReferencesBase
Description: Base class to hold configuration references for a vectorization profile.
"""
from pydantic import BaseModel

class ConfigurationReferencesBase(BaseModel):
    """
    Base class to hold configuration references for a vectorization profile.
    """
    pass



from typing import List
from langchain_community.callbacks import get_openai_callback
from langchain_core.prompts import PromptTemplate
from langchain_core.runnables import RunnablePassthrough, RunnableLambda
from langchain_core.output_parsers import StrOutputParser
from foundationallm.langchain.agents import LangChainAgentBase
from foundationallm.langchain.exceptions import LangChainException
from foundationallm.langchain.retrievers import RetrieverFactory, CitationRetrievalBase
from foundationallm.models.orchestration import (
    CompletionResponse
)
from foundationallm.models.agents import KnowledgeManagementCompletionRequest
from foundationallm.models.resource_providers.vectorization import (
    AzureAISearchIndexingProfile,
    AzureOpenAIEmbeddingProfile
)

class LangChainKnowledgeManagementAgent(LangChainAgentBase):
    """
    The LangChain Knowledge Management agent.
    """
    
    def invoke(self, request: KnowledgeManagementCompletionRequest) -> CompletionResponse:
        """
        Executes a completion request by querying the vector index with the user prompt.

        Parameters
        ----------
        request : KnowledgeManagementCompletionRequest
            The completion request to execute.
        
        Returns
        -------
        CompletionResponse
            Returns a CompletionResponse with the generated summary, the user_prompt,
            generated full prompt with context and token utilization and execution cost details.
        """
        self._validate_request(request)
        agent = request.agent

        prompt = self._get_prompt_from_object_id(agent.prompt_object_id, agent.orchestration_settings.agent_parameters)
        
        with get_openai_callback() as cb:
            try:
                prompt_builder = ''

                # Add the prefix, if it exists.
                if prompt.prefix is not None:
                    prompt_builder = f'{prompt.prefix}\n\n'

                # Add the message history, if it exists.
                conversation_history = agent.conversation_history
                if conversation_history is not None and conversation_history.enabled:
                    prompt_builder += self._build_conversation_history(
                        request.message_history,
                        conversation_history.max_history)

                # Insert the context into the template.
                prompt_builder += '{context}'   

                # Add the suffix, if it exists.
                if prompt.suffix is not None:
                    prompt_builder += f'\n\n{prompt.suffix}'

                # Get the vector document retriever, if it exists.
                retriever = None
                if request.agent.vectorization is not None:
                    indexing_profile = AzureAISearchIndexingProfile.from_object(
                        agent.orchestration_settings.agent_parameters[
                            agent.vectorization.indexing_profile_object_id])

                    text_embedding_profile = AzureOpenAIEmbeddingProfile.from_object(
                        agent.orchestration_settings.agent_parameters[
                            agent.vectorization.text_embedding_profile_object_id])

                    if (indexing_profile is not None) and (text_embedding_profile is not None):
                        retriever_factory = RetrieverFactory(
                                        indexing_profile,
                                        text_embedding_profile,
                                        self.config,
                                        request.settings)
                        retriever = retriever_factory.get_retriever()

                # Insert the user prompt into the template.
                if retriever is not None:    
                    prompt_builder += "\n\nQuestion: {question}"

                # Create the prompt template.
                prompt_template = PromptTemplate.from_template(prompt_builder)

                if retriever is not None:
                    chain_context = { "context": retriever | retriever.format_docs, "question": RunnablePassthrough() }
                else:
                    chain_context = { "context": RunnablePassthrough() }

                # Compose LCEL chain
                chain = (
                    chain_context
                    | prompt_template
                    | RunnableLambda(self._record_full_prompt)
                    | self._get_language_model(agent.orchestration_settings, request.settings)
                    | StrOutputParser()
                )

                completion = chain.invoke(request.user_prompt)
                citations = []
                if isinstance(retriever, CitationRetrievalBase):
                    citations = retriever.get_document_citations()
                    
                return CompletionResponse(
                    completion = completion,
                    citations = citations,
                    user_prompt = request.user_prompt,
                    full_prompt = self.full_prompt.text,
                    completion_tokens = cb.completion_tokens,
                    prompt_tokens = cb.prompt_tokens,
                    total_tokens = cb.total_tokens,
                    total_cost = cb.total_cost
                )
            except Exception as e:
                raise LangChainException(f"An unexpected exception occurred when executing the completion request: {str(e)}", 500) 



"""
Class: AzureAISearchConfigurationReferences
Description:
    Configuration references for an Azure AI Search indexing profile.
"""
from typing import Optional
from foundationallm.models.resource_providers.vectorization import ConfigurationReferencesBase

class AzureAISearchConfigurationReferences(ConfigurationReferencesBase):
    """
    Configuration references for an Azure AI Search indexing profile.
    """
    authentication_type: Optional[str] = "FoundationaLLM:Vectorization:AzureAISearchIndexingService:AuthenticationType"  
    endpoint: Optional[str] = "FoundationaLLM:Vectorization:AzureAISearchIndexingService:Endpoint"



"""
Classes:
    AzureAISearchSettings: Settings for an Azure AI Search indexing profile.
    AzureAISearchConfigurationReferences: Configuration references for an Azure AI Search indexing profile.
    AzureAISearchIndexingProfile: An Azure AI Search indexing profile.
Description:
    Settings specific to the Azure AI Search Indexes
"""
from typing import Any, Self
from foundationallm.models.resource_providers.vectorization import IndexingProfileBase
from .azure_ai_search_settings import AzureAISearchSettings
from .azure_ai_search_configuration_references import AzureAISearchConfigurationReferences
from foundationallm.models.utils import ObjectUtils
from foundationallm.langchain.exceptions import LangChainException

class AzureAISearchIndexingProfile(IndexingProfileBase):
    """
    An Azure AI Search indexing profile.
    """
    settings: AzureAISearchSettings
    configuration_references: AzureAISearchConfigurationReferences

    @staticmethod
    def from_object(obj: Any) -> Self:

        indexing_profile: AzureAISearchIndexingProfile = None
        
        try:
            indexing_profile = AzureAISearchIndexingProfile(**ObjectUtils.translate_keys(obj))
        except Exception as e:
            raise LangChainException(f"The indexing profile object provided in the agent parameters is invalid. {str(e)}", 400)
        
        if indexing_profile is None:
            raise LangChainException("The indexing object is missing in the agent parameters.", 400)

        return indexing_profile



"""
Class: Prompt
Description: Prompt model from resource provider.
"""
from typing import Optional
from pydantic import BaseModel

class MultipartPrompt(BaseModel):
    """
    Encapsulates the prompt model from resource provider.
    """
    name: str
    type: Optional[str] = None
    object_id: Optional[str] = None
    description: str
    prefix: Optional[str] = ""
    suffix: Optional[str] = "" 



from typing import List
from abc import ABC, abstractmethod
from foundationallm.models.orchestration import Citation

class CitationRetrievalBase(ABC):
    """
    Abstract base class indicating the ability for a retriever to retrieve citations.
    """
    @abstractmethod
    def get_document_citations(self) -> List[Citation]:
        """
        Gets sources from documents retrieved from the retriever.
        
        Returns:
            List of citations from the retrieved documents.
        """



import uuid
from typing import List, Optional
from pydantic import BaseModel
from foundationallm.models.orchestration import MessageHistoryItem, OrchestrationSettings

class CompletionRequestBase(BaseModel):
    """
    Orchestration completion request.
    """
    request_id: Optional[str] = str(uuid.uuid4())
    session_id: Optional[str] = None
    user_prompt: str    
    message_history: Optional[List[MessageHistoryItem]] = []
    settings: Optional[OrchestrationSettings] = None



"""FoundationaLLM orchestration models module"""
from .message_history_item import MessageHistoryItem
from .citation import Citation
from .orchestration_settings import OrchestrationSettings
from .completion_request_base import CompletionRequestBase
from .completion_response import CompletionResponse
from .endpoint_settings import EndpointSettings
from .operation_types import OperationTypes



from foundationallm.config import Configuration, Context
from foundationallm.langchain.agents import AgentFactory, LangChainAgentBase
from foundationallm.models.orchestration import (
    CompletionRequestBase,
    CompletionResponse
)

class OrchestrationManager:
    """Client that acts as the entry point for interacting with the FoundationaLLM Python SDK."""

    def __init__(self,
                 completion_request: CompletionRequestBase,
                 configuration: Configuration,
                 context: Context):
        """
        Initializes an instance of the OrchestrationManager.
        
        Parameters
        ----------
        completion_request : CompletionRequest
            The CompletionRequest is the metadata object containing the details needed for
            the OrchestrationManager to assemble an agent with a language model and data source.
        context : Context
            The user context under which to execution completion requests.
        """
        self.agent = self.__create_agent(
            completion_request = completion_request,
            config = configuration,
            context = context
        )

    def __create_agent(
            self,
            config: Configuration,
            completion_request: CompletionRequestBase,
            context: Context) -> LangChainAgentBase:
        """Creates an agent for executing completion requests."""
        agent_factory = AgentFactory(
            completion_request=completion_request,
            config=config,
            context=context
        )
        return agent_factory.get_agent()

    def invoke(self, request: CompletionRequestBase) -> CompletionResponse:
        """
        Executes a completion request against the LanguageModel using 
        the LangChain agent assembled by the OrchestrationManager.
        
        Parameters
        ----------
        prompt : str
            The prompt for which a completion is being generated.
            
        Returns
        -------
        CompletionResponse
            Object containing the completion response and token usage details.
        """
        return self.agent.invoke(request)



from typing import Optional
from azure.identity import DefaultAzureCredential
from langchain_core.retrievers import BaseRetriever
from foundationallm.config import Configuration
from foundationallm.langchain.language_models.openai import OpenAIModel
from foundationallm.models.orchestration import OrchestrationSettings
from foundationallm.models.language_models import EmbeddingModel, LanguageModelType, LanguageModelProvider
from .agent_parameter_retriever_keys import FILTERS, TOP_N
from .azure_ai_search_service_retriever import AzureAISearchServiceRetriever
from foundationallm.models.resource_providers.vectorization import (
    AzureAISearchIndexingProfile,
    AzureOpenAIEmbeddingProfile
)

class RetrieverFactory:
    """
    Factory class for determine which retriever to use.
    """
    def __init__(
                self,
                indexing_profile: AzureAISearchIndexingProfile,
                text_embedding_profile:AzureOpenAIEmbeddingProfile,
                config: Configuration,
                settings: Optional[OrchestrationSettings] = None
                ):
        self.config = config
        self.indexing_profile = indexing_profile
        self.text_embedding_profile = text_embedding_profile
        self.orchestration_settings = settings       

    def get_retriever(self) -> BaseRetriever:
        """
        Retrieves the retriever to use for completion requests.
        
        Returns
        -------
        BaseRetriever
            Returns the concrete initialization of a vectorstore retriever.
        """

        # use embedding profile to build the embedding model (currently only supporting Azure OpenAI)         
        #embedding_model_type = self.text_embedding_profile["text_embedding"]
        #embedding_model = None
        #match embedding_model_type:            
        #    case "SemanticKernelTextEmbedding": # same as Azure Open AI Embedding        
        e_model = EmbeddingModel(
            type = LanguageModelType.OPENAI,
            provider = LanguageModelProvider.MICROSOFT,
            # the OpenAI model uses config to retrieve the app config values - pass in the keys
            deployment = self.text_embedding_profile.configuration_references.deployment_name,
            api_endpoint = self.text_embedding_profile.configuration_references.endpoint,
            api_key = self.text_embedding_profile.configuration_references.api_key,
            api_version = self.text_embedding_profile.configuration_references.api_version
        )
        oai_model = OpenAIModel(config = self.config)
        embedding_model = oai_model.get_embedding_model(e_model)
                  
        # use indexing profile to build the retriever (current only supporting Azure AI Search)
        #vector_store_type = self.indexing_profile["indexer"]        
        #match vector_store_type:
        #    case "AzureAISearchIndexer":
        
        credential_type = self.config.get_value(self.indexing_profile.configuration_references.authentication_type)
        credential = None
        if credential_type == "AzureIdentity":            
            credential = DefaultAzureCredential(exclude_environment_credential=True)
        # NOTE: Support for all other authentication types has been removed.

        # defaults for agent parameters
        top_n = self.indexing_profile.settings.top_n
        filters = self.indexing_profile.settings.filters
        # check for settings override       
        if self.orchestration_settings is not None:
            if self.orchestration_settings.agent_parameters is not None:
                if TOP_N in self.orchestration_settings.agent_parameters:
                    top_n = self.orchestration_settings.agent_parameters[TOP_N]                    
                if FILTERS in self.orchestration_settings.agent_parameters:
                    filters = self.orchestration_settings.agent_parameters[FILTERS]                    

        retriever = AzureAISearchServiceRetriever( 
            endpoint = self.config.get_value(self.indexing_profile.configuration_references.endpoint),
            index_name = self.indexing_profile.settings.index_name,
            top_n = top_n,
            embedding_field_name = self.indexing_profile.settings.embedding_field_name,
            text_field_name = self.indexing_profile.settings.text_field_name,
            id_field_name = self.indexing_profile.settings.id_field_name,
            metadata_field_name = self.indexing_profile.settings.metadata_field_name,
            filters = filters,
            credential = credential,            
            embedding_model = embedding_model
        )
        return retriever



from typing import List, Literal
from foundationallm.langchain.data_sources import DataSourceConfiguration
from foundationallm.langchain.data_sources.search_service import SearchServiceConfiguration

class CXOConfiguration(SearchServiceConfiguration):

    """Configuration settings for a stock agent"""
    configuration_type: Literal['cxo']
    sources : List[str]
    retriever_mode : str
    company : str



from abc import abstractmethod

from langchain_community.utilities.sql_database import SQLDatabase

from foundationallm.config import Configuration
from foundationallm.langchain.data_sources.sql import SQLDatabaseConfiguration

class SQLDatabaseDataSource:
    """Data source object for a SQL database."""
    dialect: str
    driver: str
    host: str
    port: int = None
    database_name: str
    username: str
    password: str

    def __init__(self, sql_db_config: SQLDatabaseConfiguration, config: Configuration):
        """
        Initializes a SQL database data source.

        Parameters
        ----------
        sql_db_config : SqlDbConfig
            SQL Database configuration class for providing settings to the database.
        config : Configuration
            Application configuration class for retrieving configuration settings.
        """
        self.sql_db_config = sql_db_config
        self.dialect = sql_db_config.dialect
        self.driver = self.get_driver()
        self.host = sql_db_config.host
        self.port = sql_db_config.port or self.get_default_port()
        self.database_name = sql_db_config.database_name
        self.username = sql_db_config.username
        self.password = config.get_value(self.sql_db_config.password_secret_setting_key_name)
        self.include_tables = sql_db_config.include_tables
        self.sample_rows_in_table_info = 0 if sql_db_config.row_level_security_enabled \
                                            else sql_db_config.few_shot_example_count
        self.row_level_security_enabled = sql_db_config.row_level_security_enabled

    def get_database(self) -> SQLDatabase:
        """
        Retrieves a SQLDatabase object
        
        Returns
        -------
        SQLDatabase
            Returns a connection to a SQL database.
        """
        return SQLDatabase.from_uri(
            database_uri = self.get_connection_string(),
            include_tables=self.include_tables,
            sample_rows_in_table_info=self.sample_rows_in_table_info
        )

    @abstractmethod
    def get_connection_string(self) -> str:
        """retrieves the correct connection string for a given database type"""

    @abstractmethod
    def get_driver(self) -> str:
        """Retrieves the driver to use for connecting to the database."""

    @abstractmethod
    def get_default_port(self) -> int:
        """Retrieves the default port for the database."""



"""
Class name: SearchServiceConfiguration
Description: Azure AI Search index information.
"""
from typing import Optional, Literal
from foundationallm.langchain.data_sources import DataSourceConfiguration

class SearchServiceConfiguration(DataSourceConfiguration):
    """
    Connection information indicating the index information
    and the number of results to return in the RAG pattern
    """
    configuration_type: Literal['search_service']
    endpoint: str
    key_secret: str
    index_name: str
    embedding_field_name: Optional[str] = "Embedding"
    text_field_name: Optional[str] = "Text"
    top_n: Optional[int] = 3



from .settings_base import SettingsBase
from .configuration_references_base import ConfigurationReferencesBase
from .profile_base import ProfileBase

from .indexing_profiles.indexing_profile_base import IndexingProfileBase
from .indexing_profiles.azure_ai_search.azure_ai_search_settings import AzureAISearchSettings
from .indexing_profiles.azure_ai_search.azure_ai_search_configuration_references import AzureAISearchConfigurationReferences
from .indexing_profiles.azure_ai_search.azure_ai_search_indexing_profile import AzureAISearchIndexingProfile

from .embedding_profiles.embedding_profile_base import EmbeddingProfileBase
from .embedding_profiles.azure_openai.azure_openai_configuration_references import AzureOpenAIConfigurationReferences
from .embedding_profiles.azure_openai.azure_openai_embedding_profile import AzureOpenAIEmbeddingProfile



import re

class ObjectUtils:

    @staticmethod
    def pascal_to_snake(name):  
        # Convert PascalCase or camelCase to snake_case  
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)  
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()  

    @staticmethod
    def translate_keys(obj):  
        if isinstance(obj, dict):  
            new_dict = {}  
            for key, value in obj.items():  
                new_key = ObjectUtils.pascal_to_snake(key)  
                new_dict[new_key] = ObjectUtils.translate_keys(value)  # Recursively apply to values  
            return new_dict  
        elif isinstance(obj, list):  
            return [ObjectUtils.translate_keys(item) for item in obj]  # Apply to each item in the list  
        else:  
            return obj  # Return the item itself if it's not a dict or list



from langchain_core.language_models import BaseLanguageModel
from foundationallm.config import Configuration
from foundationallm.langchain.language_models.openai import OpenAIModel
from foundationallm.models.orchestration import OrchestrationSettings

class LanguageModelFactory:
    """
    Factory class for determine which language models to use.
    """
    def __init__(self, orchestration_settings: OrchestrationSettings, config: Configuration):
        self.config = config

    def get_llm(self) -> BaseLanguageModel:
        """
        Retrieves the language model to use for completion and embedding requests.
        
        Returns
        -------
        BaseLanguageModel
            Returns the language model to use for completion and embedding requests.
        """
        return OpenAIModel(config = self.config)



from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseLanguageModel
from langchain_openai import AzureOpenAI, OpenAI, AzureChatOpenAI, ChatOpenAI, AzureOpenAIEmbeddings, OpenAIEmbeddings

from foundationallm.config import Configuration
from foundationallm.langchain.language_models import LanguageModelBase
from foundationallm.models.language_models import (
    EmbeddingModel,
    LanguageModelProvider
)

class OpenAIModel(LanguageModelBase):
    """OpenAI Completion model."""
    config_value_base_name: str

    def __init__(self, config: Configuration):
        """
        Initializes an OpenAI completion language model.
        
        Parameters
        ----------
        language_model: LanguageModel
            The language model metadata class.
        config : Configuration
            Application configuration class for retrieving configuration settings.
        """
        self.config = config

    # def get_completion_model(self, language_model: LanguageModel) -> BaseLanguageModel:
    #     """
    #     Returns an OpenAI completion model.
        
    #     Returns
    #     -------
    #     BaseLanguageModel
    #         Returns an OpenAI completion model.
    #     """
    #     use_chat = language_model.use_chat

    #     openai_base_url = self.config.get_value(language_model.api_endpoint)
    #     openai_api_key = self.config.get_value(language_model.api_key)
    #     openai_api_version = self.config.get_value(language_model.api_version)

    #     # Overridable values
    #     azure_deployment_name = language_model.deployment
    #     temperature = language_model.temperature or 0.0

    #     if language_model.provider == LanguageModelProvider.MICROSOFT:        
    #         if use_chat:
    #             return AzureChatOpenAI(
    #                 api_key = openai_api_key,
    #                 api_version = openai_api_version,
    #                 azure_deployment = azure_deployment_name,
    #                 azure_endpoint = openai_base_url,
    #                 #max_tokens = self.config.get_value(f'{config_value_base_name}:Completions:MaxTokens'),
    #                 model_version = self.config.get_value(language_model.version),
    #                 temperature = temperature
    #             )
    #         else:
    #             return AzureOpenAI(
    #                 api_key = openai_api_key,
    #                 api_version = openai_api_version,
    #                 azure_deployment = azure_deployment_name,
    #                 azure_endpoint = openai_base_url,
    #                 #max_tokens = self.config.get_value(f'{config_value_base_name}:Completions:MaxTokens'),
    #                 temperature = temperature
    #             )
    #     else:
    #         if use_chat:
    #             return ChatOpenAI(
    #                 base_url = openai_base_url,
    #                 api_key = openai_api_key,
    #                 temperature = temperature
    #             )
    #         else:
    #             return OpenAI(
    #                 base_url = openai_base_url,
    #                 api_key = openai_api_key,
    #                 temperature = temperature
    #             )

    def get_embedding_model(self, embedding_model: EmbeddingModel) -> Embeddings:
        """
        Retrieves the OpenAI embedding model.
        
        Returns
        -------
        Embeddings
            Returns an OpenAI embeddings model.
        """
        if embedding_model is None:
            raise ValueError('Expected populated embedding_model, got None.')

        if embedding_model.provider == LanguageModelProvider.MICROSOFT:
            return AzureOpenAIEmbeddings(
                openai_api_key = self.config.get_value(embedding_model.api_key),
                openai_api_version = self.config.get_value(embedding_model.api_version),
                deployment = self.config.get_value(embedding_model.deployment),
                azure_endpoint = self.config.get_value(embedding_model.api_endpoint),
                chunk_size = embedding_model.chunk_size
            )
        else:
            return OpenAIEmbeddings(
                api_key = self.config.get_value(embedding_model.api_key),
                api_version = self.config.get_value(embedding_model.api_version),
                base_url = self.config.get_value(embedding_model.api_endpoint),
                chunk_size = embedding_model.chunk_size or 1000,
                deployment = self.config.get_value(embedding_model.deployment),
                model = self.config.get_value(embedding_model.model)
            )



"""
Encapsulates properties agent vectorization settings.
"""
from typing import Optional
from pydantic import BaseModel

class AgentVectorizationSettings(BaseModel):
    """
    Encapsulates properties for agent vectorization settings.
    """
    dedicated_pipeline: Optional[bool] = False
    data_source_object_id: Optional[str] = None
    indexing_profile_object_id: Optional[str] = None
    text_embedding_profile_object_id: Optional[str] = None
    text_partitioning_profile_object_id: Optional[str] = None
    vectorization_data_pipeline_object_id: Optional[str] = None



"""
Class: AzureOpenAIConfigurationReferences
Description:
    Configuration references for an Azure OpenAI text embedding service.
"""
from typing import Optional
from foundationallm.models.resource_providers.vectorization import ConfigurationReferencesBase

class AzureOpenAIConfigurationReferences(ConfigurationReferencesBase):
    """
    Configuration references for an Azure OpenAI text embedding service.
    """
    api_key: Optional[str] = "FoundationaLLM:Vectorization:SemanticKernelTextEmbeddingService:APIKey"
    api_version: Optional[str] = "FoundationaLLM:Vectorization:SemanticKernelTextEmbeddingService:APIVersion"
    authentication_type: Optional[str] = "FoundationaLLM:Vectorization:SemanticKernelTextEmbeddingService:AuthenticationType"
    deployment_name: Optional[str] = "FoundationaLLM:Vectorization:SemanticKernelTextEmbeddingService:DeploymentName"
    endpoint: Optional[str] = "FoundationaLLM:Vectorization:SemanticKernelTextEmbeddingService:Endpoint"
